#!/usr/bin/python3
import csv
import math
import cmath
import time
import numpy
from argparse import ArgumentParser
import pandas as pd
import matplotlib.pyplot as plt
import climate_chamber
import VNA
import virtual_time
import json
import MeasurementPlot
import sys
import analysis

TEMPERATURE_STABLE = 0x1
HUMIDITY_STABLE = 0x2
MAGNITUDE_STABLE = 0x4
PHASE_STABLE = 0x8
MEASUREMENT_STABLE = 0x10


# Class has only attributes which we are using in the read_data_function to read data from VNA and Chamber
class MeasurementData:
    def __init__(self, timestamp, temp, hum, power, frequency, s11, s21, s12, s22):
        self.timestamp = timestamp
        self.temp = temp
        self.hum = hum
        self.power = power
        self.frequency = frequency
        self.s11 = s11
        self.s21 = s21
        self.s12 = s12
        self.s22 = s22


class Measurements:
    def __init__(self, chamber_address, vna_address, output_basename, standby, config_data):
        self.max_delta_temp = config_data['delta_temp']
        self.max_delta_hum = config_data['delta_hum']
        self.max_delta_mag = config_data['delta_mag']
        self.max_delta_phase = config_data['delta_phase']
        self.sleep_time = config_data["sleep_time"]
        self.frequency = config_data["frequency"]
        self.vna_config_file = config_data["vna_config_file"]
        target_accuracy = [self.max_delta_temp, self.max_delta_hum]
        self.chamber = climate_chamber.create_chamber(chamber_address, target_accuracy)
        self.vna = VNA.create_vna(vna_address, target_accuracy)
        self.standby = standby
        self.output_basename = output_basename
        self.clock = virtual_time.get_clock(chamber_address, target_accuracy)
        self.temperature_stable = False
        self.humidity_stable = False
        self.magnitude_stable = False
        self.phase_stable = False

        self.vna.load_config(self.vna_config_file, self.frequency)
        self.vna.create_new_trace("Trace1", "S11")
        self.vna.create_new_trace("Trace2", "S12")
        self.vna.create_new_trace("Trace3", "S21")
        self.vna.create_new_trace("Trace4", "S22")

        self.measurement_plot = MeasurementPlot.MeasurementPlot()
        self.data_collection = []

    def perform_measurements(self, sweep_file):
        with open(sweep_file) as file:
            try:
                measurement_number = 0
                while line := file.readline().rstrip():
                    list_of_values = line.split()
                    next_temp = float(list_of_values[0])  # target_temperature
                    next_hum = float(list_of_values[1])  # target_humidity
                    next_soaking = int(list_of_values[2])  # soaking_time
                    next_reads = int(list_of_values[3])  # number of stable readings

                    self.measurement_plot.fig.suptitle("Measurement "+str(measurement_number)+': ' + str(next_temp) +
                                                       ' degC, ' + str(next_hum) + ' rel. hum.', color="red")
                    self.perform_single_measurement(self.output_basename+'_'+str(measurement_number)+'.csv', next_temp, next_hum,
                                                    next_soaking, next_reads)
                    measurement_number += 1

            except KeyboardInterrupt:
                pass

        plt.close()

        if self.standby:
            standby_response = self.chamber.set_mode('STANDBY')
            print(standby_response)

        return measurement_number

    def perform_temperature_sweep(self, start_temp, stop_temp, temp_step, target_hum, soaking_time, n_stable_reads):
        current_temp = start_temp
        temperatures = []
        while current_temp < stop_temp:
            temperatures.append(current_temp)
            current_temp = round(current_temp + temp_step, 1)
        temperatures.append(stop_temp)

        for temp in temperatures:
            output_name = self.output_basename + '_'+str(temp)+'deg_'+str(target_hum)+'rh.csv'
            self.measurement_plot.fig.suptitle("Temp. sweep measurement "+str(temperatures.index(temp)+1)+'/'+str(len(temperatures))+
                                               ': ' + str(temp) + ' degC, ' + str(target_hum) + ' rel. hum.',
                                               color="red")
            self.perform_single_measurement(output_name, temp, target_hum, soaking_time, n_stable_reads)

        plt.close()
        return temperatures

    def perform_single_measurement(self, output, target_temp, target_hum, soaking_time, n_stable_reads):
        with open(output, mode='w', newline='') as csv_file:
            fieldnames = ['TIMESTAMP', 'TARGET_TEMPERATURE', 'READBACK_TEMPERATURE', 'TARGET_HUMIDITY',
                          'READBACK_HUMIDITY', 'RF_POWER', 'RF_FREQUENCY', 'DUT_IDENTIFIER', 'RUN_ID',
                          'EQUILIBRIUM_INDICATOR', 'S11_MAGNITUDE', 'S11_PHASE', 'S12_MAGNITUDE', 'S12_PHASE',
                          'S21_MAGNITUDE', 'S21_PHASE', 'S22_MAGNITUDE', 'S22_PHASE']
            # csv.dict writer add adda row wise
            writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
            writer.writeheader()
            self.data_collection = []
            plt.ion()

            set_const_response = self.chamber.set_const((target_temp, target_hum))
            print(set_const_response)
            set_mode_response = self.chamber.set_mode('CONSTANT')
            print(set_mode_response)

            number_of_soaking_reads = soaking_time / self.sleep_time + 1

            do_another_measurement = True
            #next_read_time is the starttime of the second read (i.e. read after this one). The time of
            #this read (i.e. the first read in the measurement) is now().
            next_read_time = self.clock.time() + self.sleep_time
            while do_another_measurement:
                # wait until set point is reached (+soaking time)
                magnitudes_queue = []
                phase_queue = []

                while True:
                    data = self.read_data()
                    # if it is not within the target range reset start time
                    self.temperature_stable = False
                    self.humidity_stable = False

                    self.temperature_stable = self.calculate_temperature_stability(target_temp, float(data.temp))
                    self.humidity_stable = self.calculate_humidity_stability(target_hum, float(data.hum))

                    # The queue must not be longer than the max number of soaking reads.
                    # If the queue is already full, we have to pop the first element before we can add the
                    # current measurement.
                    if len(magnitudes_queue) >= number_of_soaking_reads:
                        magnitudes_queue.pop(0)
                    if self.temperature_stable and self.humidity_stable:
                        magnitudes_queue.append(self.calculate_mean_magnitude(data.s21))
                    else:
                        magnitudes_queue.clear()
                    # check cable stability parameters
                    self.magnitude_stable = False
                    if len(magnitudes_queue) >= number_of_soaking_reads:
                        spread = max(magnitudes_queue) - min(magnitudes_queue)
                        if spread < 2*self.max_delta_mag:
                            self.magnitude_stable = True

                    if len(phase_queue) >= number_of_soaking_reads:
                        phase_queue.pop(0)
                    if self.temperature_stable and self.humidity_stable:
                        phase_queue.append(self.calculate_mean_phase(data.s21))
                    else:
                        phase_queue.clear()

                    self.phase_stable = False
                    if len(phase_queue) >= number_of_soaking_reads:
                        spread = max(phase_queue) - min(phase_queue)
                        if spread < 2*self.max_delta_phase:
                            self.phase_stable = True

                    print('Setpoint: ' + str(target_temp) + ' ' + str(target_hum) + ' | Temp: ' + data.temp +
                          ' Â°C' + ' | Humid: ' + data.hum + '%'
                          + ' | soaking read nr' + str(len(magnitudes_queue)))
                    self.store_and_plot_data(target_temp, target_hum, data, self.cook_up_equi_indicator())
                    writer.writerow(self.data_collection[-1])

                    if self.temperature_stable and self.humidity_stable and self.magnitude_stable and\
                            self.phase_stable:
                        reference_magnitude = magnitudes_queue[-1]
                        reference_phase = phase_queue[-1]
                        print('SOAKING FINISHED!')
                        break
                    else:
                        self.sleep_until(next_read_time)
                        next_read_time += self.sleep_time

                # perform the configured number of measurements and check that they are really stable
                # It started running after everything become stable
                supposedly_stable_measurements = []
                all_measurements_stable = True
                for i in range(0, n_stable_reads):
                    data = self.read_data()
                    self.temperature_stable = self.calculate_temperature_stability(target_temp, float(data.temp))
                    self.humidity_stable = self.calculate_humidity_stability(target_hum, float(data.hum))
                    mag = self.calculate_mean_magnitude(data.s21)
                    phase = self.calculate_mean_phase(data.s21)
                    self.magnitude_stable = (reference_magnitude-self.max_delta_mag <= mag) and\
                                            (mag <= reference_magnitude+self.max_delta_mag)
                    self.phase_stable = (reference_phase-self.max_delta_phase <= phase) and\
                                        (phase <= reference_phase+self.max_delta_phase)

                    self.store_and_plot_data(target_temp, target_hum, data, self.cook_up_equi_indicator())
                    supposedly_stable_measurements.append(self.data_collection[-1])
                                
                    if (self.temperature_stable and self.humidity_stable and self.magnitude_stable and
                            self.phase_stable):
                        print('Stable measurement ' + str(i+1) + '/' + str(n_stable_reads))
                        self.sleep_until(next_read_time)
                        next_read_time += self.sleep_time
                    else:
                        all_measurements_stable = False
                        print('Measurement not stable. Retrying.')
                        break

                for measurement in supposedly_stable_measurements:
                    if all_measurements_stable:
                        measurement['EQUILIBRIUM_INDICATOR'] = TEMPERATURE_STABLE | HUMIDITY_STABLE |\
                                                               MAGNITUDE_STABLE | PHASE_STABLE |\
                                                               MEASUREMENT_STABLE
                        do_another_measurement = False

                    writer.writerow(measurement)

    def sleep_until(self, wakeup_time):
        remaining_sleep_time = wakeup_time - self.clock.time()
        if remaining_sleep_time > 0:
            self.clock.sleep(remaining_sleep_time)

    def read_data(self):
        [temp, hum, mode, alarms] = self.chamber.read_monitor().split(',')
        power = self.vna.get_current_power()
        frequency = self.vna.get_current_cw_frequency()
        self.vna.do_single_sweep()
        s11 = self.get_trace_data("Trace1")
        s12 = self.get_trace_data("Trace2")
        s21 = self.get_trace_data("Trace3")
        s22 = self.get_trace_data("Trace4")

        return MeasurementData(int(self.clock.time()), temp, hum, power, frequency, s11, s21, s12, s22)

    def store_and_plot_data(self, target_temp, target_hum, data, equi_indicator):
        measurement = {
            'TIMESTAMP': data.timestamp,
            'TARGET_TEMPERATURE': target_temp,
            'READBACK_TEMPERATURE': float(data.temp),
            'TARGET_HUMIDITY': target_hum,
            'READBACK_HUMIDITY': float(data.hum),
            'RF_POWER': data.power,
            'RF_FREQUENCY': data.frequency,
            'EQUILIBRIUM_INDICATOR': equi_indicator,
            'S11_PHASE': self.calculate_mean_phase(data.s11),
            'S11_MAGNITUDE': self.calculate_mean_magnitude(data.s11),
            'S21_PHASE': self.calculate_mean_phase(data.s21),
            'S21_MAGNITUDE': self.calculate_mean_magnitude(data.s21),
            'S12_PHASE': self.calculate_mean_phase(data.s12),
            'S12_MAGNITUDE': self.calculate_mean_magnitude(data.s12),
            'S22_PHASE': self.calculate_mean_phase(data.s22),
            'S22_MAGNITUDE': self.calculate_mean_magnitude(data.s22)
        }
        self.data_collection.append(measurement)
        data_frame = pd.DataFrame(self.data_collection)
        self.measurement_plot.draw(data_frame)

    def current_milli_time(self):
        return int(round(self.clock.time() * 1000))

    def get_trace_data(self, trace):
        return self.vna.get_list_of_measurement_values(trace, "SDAT")

    def calculate_complex_numbers(self, values_list):
        real_num = values_list[::2]
        imaginary_num = values_list[1::2]
        complex_numbers = []
        for i, q in zip(real_num, imaginary_num):
            complex_numbers.append(complex(i, q))
        return complex_numbers

    def calculate_magnitudes(self, values_list):
        complex_numbers = self.calculate_complex_numbers(values_list)
        magnitudes = [abs(val) for val in complex_numbers]
        return magnitudes

    def calculate_mean_magnitude(self, values_list):
        magnitudes = self.calculate_magnitudes(values_list)
        return numpy.mean(magnitudes)

    def calculate_phases(self, values_list):
        complex_numbers = self.calculate_complex_numbers(values_list)
        phases = [math.degrees(cmath.phase(val)) for val in complex_numbers]
        return phases

    def calculate_mean_phase(self, values_list):
        phases = self.calculate_phases(values_list)
        return numpy.mean(phases)

    def cook_up_equi_indicator(self):
        equilibrium_indicator = 0

        if self.temperature_stable:
            equilibrium_indicator = equilibrium_indicator | TEMPERATURE_STABLE

        if self.humidity_stable:
            equilibrium_indicator = equilibrium_indicator | HUMIDITY_STABLE

        if self.magnitude_stable:
            equilibrium_indicator = equilibrium_indicator | MAGNITUDE_STABLE

        if self.phase_stable:
            equilibrium_indicator = equilibrium_indicator | PHASE_STABLE

        return equilibrium_indicator

    def calculate_temperature_stability(self, target_temp, readback_temp):
        return (target_temp-self.max_delta_temp <= float(readback_temp)) and \
               (float(readback_temp) <= target_temp+self.max_delta_temp)

    def calculate_humidity_stability(self, target_hum, readback_hum):
        return (target_hum-self.max_delta_hum <= float(readback_hum)) and \
               (float(readback_hum) <= target_hum+self.max_delta_hum)

def plot_output(output_basename, measurements_appendices, show_blocking_plot, title = ''):
    list_of_frames = []
    for m in measurements_appendices:
        measurement_name = output_basename+'_'+str(m)
        list_of_frames.append(pd.read_csv(measurement_name+'.csv'))

    combined_data_frame = pd.concat(list_of_frames, ignore_index=True, sort=False)
    if show_blocking_plot:
        plt.ioff()
    plot = MeasurementPlot.MeasurementPlot(title)
    plot.draw(combined_data_frame, output_basename + '_graph.pdf')

def run_temperature_sweep_from_file(temperature_sweep_file, meas):
    with open(temperature_sweep_file) as file:
        try:
            line = file.readline().rstrip()
            list_of_values = line.split()
            start_temp = float(list_of_values[0])
            stop_temp = float(list_of_values[1])
            temperature_step = float(list_of_values[2])
            target_hum = float(list_of_values[3])
            soaking_time = int(list_of_values[4])
            n_stable_reads = int(list_of_values[5])

            return meas.perform_temperature_sweep(start_temp, stop_temp, temperature_step, target_hum, soaking_time,
                                                  n_stable_reads), target_hum

        except KeyboardInterrupt:
            pass

if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument("-c", "--chamber",
                        help="IP address of climate chamber", metavar="ADDR",
                        required=True)
    parser.add_argument("-v", "--vna",
                        help="IP address of VNA", metavar="ADDR",
                        required=True)
    parser.add_argument('-f', '--file',
                        help='File containing custom list of measurements',
                        default='')
    parser.add_argument('-t', '--temperaturesweepfile',
                        help='File containing temperature sweep parameters',
                        default='')
    parser.add_argument('-o', '--output',
                        help='Name of output file')
    parser.add_argument("-s", "--standby",
                        action="store_true", default=False,
                        help="Change Chamber mode to standby at the end")
    parser.add_argument("-p", "--plot",
                        action="store_true", default=False,
                        help="Show blocking plot window at the end")

    # parser.add_argument("-q", "--quiet",
    #                     action="store_false", dest="verbose", default=True,
    #                     help="don't print status messages to stdout")
    args = parser.parse_args()

    if args.file == '' and args.temperaturesweepfile == '':
        print('Argument error: Either \'file\' or \'temperaturesweepfile\' must be specified.')
        sys.exit(-1)

    if not args.output:
        output_basename = time.strftime("%Y_%m_%d-%H_%M_%S") + "_results"
    else:
        output_basename = args.output

    print(args.chamber, args.vna, args.file, output_basename, args.standby)

    # reading json file for target accuracy
    with open('test_stand_parameter.json', 'r') as f:
        config_data = json.load(f)

    mes = Measurements(args.chamber, args.vna, output_basename, args.standby, config_data)
    try:
        if args.file:
            n_measurements = mes.perform_measurements(args.file)
            plot_output(output_basename, range(n_measurements), args.plot, output_basename)
        if args.temperaturesweepfile:
            temperatures, humidity = run_temperature_sweep_from_file(args.temperaturesweepfile, mes)
            #run analysis here
            temp_extensions = []
            for t in temperatures:
                temp_extensions.append(str(t)+'deg_'+str(humidity)+'rh')
            analysis.plot_sweep(temperatures, [humidity]*len(temperatures), output_basename, 'temperature')
            plot_output(output_basename, temp_extensions, args.plot, output_basename + ': Temperature sweep ' +
                        str(temperatures[0]) + ' degC to ' + str(temperatures[-1]) + ' degC')
            print(str(temp_extensions))
    finally:
        mes.chamber.close()

    # mes.plot_output(output)

